from os import path, remove, chdir, makedirs, system
from jinja2.environment import Environment
from jinja2 import FileSystemLoader, exceptions
from pykwalify.errors import PyKwalifyException
from pykwalify.core import Core
from hitchtest.utils import warn
from hitchtest.test import Test
import subprocess
import inspect
import jinja2
import yaml
import time
import sys
import imp
import sys
import os


HITCHSCHEMA = path.join(path.dirname(path.realpath(__file__)), "schema.yaml")

class Module(object):
    def __init__(self, filename, settings):
        self.filename = path.realpath(filename)
        self.engine_folder = settings['engine_folder']

        self.name = path.split(self.filename)[1].replace(".test", "")
        self.title = self.name.replace("_", " ").title()
        self.dirname = path.dirname(self.filename)


        if settings is None:
            self.settings = {}
        else:
            self.settings = settings

        with open(filename, "r") as file_handle:
            if file_handle.read() == "":
                warn("{0} is an empty file.\n".format(filename))
                sys.exit(1)

        env = Environment()
        env.loader = FileSystemLoader(path.split(filename)[0])
        try:
            template = env.get_template(path.split(filename)[1])
        except exceptions.TemplateError as error:
            warn("Jinja2 template error in '{}' on line {}:\n==> {}\n".format(
                error.filename, error.lineno, str(error)
            ))
            sys.exit(1)
        self.test_yaml_text = template.render(**self.settings.as_dict())

        if self.test_yaml_text == "":
            warn((
                "{0} rendered as an empty file. "
                "There is probably a problem with your jinja2 template.\n"
            ).format(filename))
            sys.exit(1)

        self.tests = []
        try:
            module_yaml_as_dict = yaml.load(self.test_yaml_text)
        except yaml.parser.MarkedYAMLError as error:
            warn("YAML parser error in {}:\n".format(filename))
            warn(str(error))
            warn("\n")
            sys.exit(1)

        try:
            core = Core(source_data=module_yaml_as_dict, schema_files=[HITCHSCHEMA])
            core.validate(raise_exception=True)
        except PyKwalifyException as error:
            warn("YAML validation error in {}:\n==> {}\n".format(filename, error.msg))
            sys.exit(1)

        if len(module_yaml_as_dict) == 1:
            self.multiple_tests = False
            self.tests = [Test(module_yaml_as_dict[0], self.settings, self.filename, self.title)]
        else:
            self.multiple_tests = True
            for i, test_yaml in enumerate(module_yaml_as_dict, 1):
                self.tests.append(
                    Test(test_yaml, self.settings, self.filename, "{} {}".format(self.title, i))
                )



from hitchtest.step import Step

class Scenario(object):
    """The series of steps that make up a hitch test."""

    def __init__(self, yaml_scenario, underscore_case_steps=True):
        self.steps = []

        if yaml_scenario is not None:
            for i, yaml_step in enumerate(yaml_scenario, 1):
                self.steps.append(Step(yaml_step, i, underscore_case_steps=underscore_case_steps))

    def to_dict(self):
        return [step.to_dict() for step in self.steps]



from os import path, getpgrp, getpgid, tcsetpgrp, fdopen
from hitchtest.utils import log, warn, signals_trigger_exit, ignore_signals, signal_pass_on_to_separate_process_group
from hitchtest.results import Results
from hitchtest.result import Result
from functools import partial
import multiprocessing
import psutil
import signal
import termios
import sys
import os


class Suite(object):
    """A group of tests defined at runtime."""
    def __init__(self, test_modules, settings, tags):
        self.test_modules = test_modules
        self.settings = settings
        self.tags = set(tags.split(",")) if tags is not None else None

    def tests(self):
        """Get the full list of tests in the specified suite."""
        full_test_list = []
        for test_module in self.test_modules:
            full_test_list.extend(test_module.tests)
        test_list = []
        if self.tags is not None:
            for test in full_test_list:
                if test.tags is not None:
                    if set(test.tags).issuperset(self.tags):
                        test_list.append(test)
        else:
            test_list = full_test_list
        return test_list

    def printyaml(self):
        """Print the test YAML for jinja2 debugging purposes."""
        for test_module in self.test_modules:
            log("\n# {}\n{}\n".format(
                test_module.filename,
                test_module.test_yaml_text
            ))

    def run(self, quiet=False):
        """Run all tests in the defined suite of modules."""
        tests = self.tests()
        failedfast = False
        result_list = []

        for test in tests:
            if quiet:
                hijacked_stdout = sys.stdout
                hijacked_stderr = sys.stderr
                sys.stdout = open(path.join(self.settings['engine_folder'], ".hitch", "test.out"), "ab", 0)
                sys.stderr = open(path.join(self.settings['engine_folder'], ".hitch", "test.err"), "ab", 0)

            def run_test_in_separate_process(file_descriptor_stdin, result_queue):
                """Change process group, run test and return result via a queue."""
                orig_pgid = os.getpgrp()
                os.setpgrp()
                result_queue.put("pgrp")
                if not quiet:
                    sys.stdin = os.fdopen(file_descriptor_stdin)
                result = test.run()
                result_queue.put(result)
                if not quiet:
                    try:
                        os.tcsetpgrp(file_descriptor_stdin, orig_pgid)
                    except OSError as error:
                        if error.args[0] == 25:
                            pass

            if not quiet:
                try:
                    orig_stdin_termios = termios.tcgetattr(sys.stdin.fileno())
                except termios.error:
                    orig_stdin_termios = None
                orig_stdin_fileno = sys.stdin.fileno()
            orig_pgid = os.getpgrp()

            file_descriptor_stdin = sys.stdin.fileno()
            result_queue = multiprocessing.Queue()


            # Start new process to run test in, to isolate it from future test runs
            test_process = multiprocessing.Process(
                target=run_test_in_separate_process,
                args=(file_descriptor_stdin, result_queue)
            )

            test_timed_out = False

            test_process.start()

            # Ignore all exit signals but pass them on
            signal_pass_on_to_separate_process_group(test_process.pid)

            # Wait until PGRP is changed
            result_queue.get()

            # Make stdin go to the test process so that you can use ipython, etc.
            if not quiet:
                try:
                    os.tcsetpgrp(file_descriptor_stdin, os.getpgid(test_process.pid))
                except OSError as error:
                    if error.args[0] == 25:
                        pass

            # Wait until process has finished
            proc = psutil.Process(test_process.pid)
            test_timeout = self.settings.get("test_timeout", None)
            test_shutdown_timeout = self.settings.get("test_shutdown_timeout", 10)

            try:
                proc.wait(timeout=test_timeout)
            except psutil.TimeoutExpired:
                test_timed_out = True
                proc.send_signal(signal.SIGTERM)

                try:
                    proc.wait(timeout=test_shutdown_timeout)
                except psutil.TimeoutExpired:
                    for child in proc.get_children(recursive=True):
                        child.send_signal(signal.SIGKILL)
                    proc.send_signal(signal.SIGKILL)


            # Take back signal handling from test running code
            signals_trigger_exit()


            try:
                result = result_queue.get_nowait()
            except multiprocessing.queues.Empty:
                result = Result(test, True, 0.0)

            if test_timed_out:
                result.aborted = False
            result_list.append(result)

            if not quiet and orig_stdin_termios is not None:
                try:
                    termios.tcsetattr(orig_stdin_fileno, termios.TCSANOW, orig_stdin_termios)
                except termios.error as err:
                    # I/O error caused by another test stopping this one
                    if err[0] == 5:
                        pass

            if quiet:
                sys.stdout = hijacked_stdout
                sys.stderr = hijacked_stderr

            if quiet and result is not None:
                if result.failure:
                    warn("X")
                else:
                    warn(".")

            if result.aborted:
                warn("Aborted\n")
                sys.exit(1)

            if self.settings.get('failfast', False) and result.failure:
                failedfast = True
                break
        return Results(result_list, failedfast, self.settings.get('colorless', False))



from hitchtest.hitch_stacktrace import HitchStacktrace, TestPosition
from hitchtest.utils import log, warn
from hitchtest.scenario import Scenario
from hitchtest.result import Result
from hitchtest.environment import verify
from os import path
import inspect
import time
import imp
import re


class Test(object):
    def __init__(self, yaml_test, settings, filename, title):
        self.settings = settings
        self.filename = filename
        self.name = yaml_test.get('name', title)
        self.engine = yaml_test.get('engine', 'engine.py:ExecutionEngine')
        self.description = yaml_test.get('description')
        self.preconditions = yaml_test.get('preconditions', {})
        self.tags = yaml_test.get('tags')
        self.features = yaml_test.get('features')
        self.scenario = Scenario(yaml_test.get('scenario', []), underscore_case_steps=self.settings.get("underscore case steps", False))

        if re.compile("^(.*?)\:(.*?)$").match(self.engine) is None:
            raise RuntimeError("ERROR : engine should be of the form 'engine_filename.py:ClassName'")
        else:
            module_source = path.abspath(path.join(self.settings['engine_folder'], self.engine.split(":")[0]))
            if not path.exists(module_source):
                raise RuntimeError("ERROR : engine filename '{}' not found.".format(self.engine.split(':')[0]))

            engine_class_name = self.engine.split(":")[1]
            engine_module = imp.load_source("engine", module_source)
            if len([x for x in inspect.getmembers(engine_module) if x[0] == engine_class_name]) == 0:
                raise RuntimeError("ERROR : Class {} not found in engine.".format(engine_class_name))
            self.engine_class = [x for x in inspect.getmembers(engine_module) if x[0] == engine_class_name][0][1]

        if self.name[0].isdigit():
            raise RuntimeError("ERROR : Test names cannot start with a digit.")
        if "test" in self.name.lower():
            warn("WARNING: The word 'test' should not appear in your test name - it is redundant.\n")

    def to_dict(self):
        return {
            "filename": self.filename,
            "name": self.name,
            "engine": self.engine,
            "preconditions": self.preconditions,
            "description": self.description,
            "tags": self.tags,
            "features": self.features,
            "scenario": self.scenario.to_dict(),
        }

    def run(self):
        start_time = time.time()
        stacktrace = None
        engine = None
        failure = False
        show_hitch_stacktrace = self.settings.get("show_hitch_stacktrace", False)

        try:
            engine = self.engine_class(self.settings, self.preconditions)
            engine.name = self.name
            engine.aborted = False
            engine.stacktrace = None
            engine._test = self
        except Exception as e:
            stacktrace = HitchStacktrace(self, TestPosition.SETUP, show_hitch_stacktrace)
            failure = True

        if not failure:
            log("RUNNING TEST {}\n".format(self.name))

            try:
                verify(self.settings.get("environment", []))
                engine.set_up()
            except Exception as e:
                stacktrace = HitchStacktrace(self, TestPosition.SETUP, show_hitch_stacktrace)
                failure = True

            if not failure:
                for step in self.scenario.steps:
                    try:
                        step.run(engine)
                    except Exception as e:
                        stacktrace = HitchStacktrace(self, TestPosition.STEP, show_hitch_stacktrace, step=step)
                        failure = True
                        break

            if not engine.aborted:
                if failure:
                    try:
                        engine.stacktrace = stacktrace
                        engine.on_failure()
                    except Exception as e:
                        stacktrace = HitchStacktrace(self, TestPosition.ON_FAILURE, show_hitch_stacktrace)
                else:
                    try:
                        engine.on_success()
                    except Exception as e:
                        failure = True
                        stacktrace = HitchStacktrace(self, TestPosition.ON_SUCCESS, show_hitch_stacktrace)

            try:
                engine.tear_down()
            except Exception as e:
                failure = True
                stacktrace = HitchStacktrace(self, TestPosition.TEARDOWN, show_hitch_stacktrace)

        duration = time.time() - start_time
        dict_stacktrace = stacktrace.to_dict() if stacktrace is not None else None
        aborted = engine.aborted if engine is not None else False
        result = Result(self, failure, duration, stacktrace=dict_stacktrace, aborted=aborted)
        return result



from IPython.terminal.embed import InteractiveShellEmbed
from traitlets.config.loader import Config
import functools
import patoolib
import IPython
import inspect
import signal
import shutil
import psutil
import os
import io


def _write(handle, message):
    if isinstance(handle, io.TextIOWrapper):
        handle.write(message)
    else:
        handle.write(message.encode('utf8'))
    handle.flush()

def log(message):
    """Output to stdout."""
    import sys
    _write(sys.stdout, message)

def warn(message):
    """Output to stderr."""
    import sys
    _write(sys.stderr, message)

def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return ''.join(x for x in text if x.isalpha() or x.isdigit() or x == "_")

def ipython(message=None, frame=None):
    """Launch into customized IPython with greedy autocompletion and no prompt to exit.
       If stdin is not tty, just issue warning message."""
    import sys
    if os.isatty(sys.stdin.fileno()):
        config = Config({
            'InteractiveShell': {'confirm_exit': False, },
            'IPCompleter': {'greedy': True, }
        })
        InteractiveShellEmbed.instance(config=config)(message, local_ns=frame.f_locals, global_ns=frame.f_globals)
    else:
        warn("==========> IPython cannot be launched if stdin is not a tty.\n\n")

def ipython_embed(message=None, functions_above=0):
    frame = inspect.stack()[functions_above + 1][0]
    ipython(message, frame)

def stop_ipython():
    from IPython.terminal.interactiveshell import TerminalInteractiveShell
    TerminalInteractiveShell.exit_now = True

def do_exit(signal, frame):
    """Just exit."""
    import sys
    sys.exit(1)

def signals_trigger_exit():
    """Make all signals cause a system exit."""
    signal.signal(signal.SIGINT, do_exit)
    signal.signal(signal.SIGTERM, do_exit)
    signal.signal(signal.SIGHUP, do_exit)
    signal.signal(signal.SIGQUIT, do_exit)

def ignore_signals():
    """Ignore all signals (e.g. ctrl-C)."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    signal.signal(signal.SIGHUP, signal.SIG_IGN)
    signal.signal(signal.SIGQUIT, signal.SIG_IGN)

def signal_pass_on_to_separate_process_group(pid):
    def pass_on_signal(pid, signum, frame):
        os.kill(pid, signum)

    signal.signal(signal.SIGINT, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGTERM, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGHUP, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGQUIT, functools.partial(pass_on_signal, pid))

def extract_archive(filename, directory):
    patoolib.extract_archive(filename, outdir=directory)


class DownloadError(Exception):
    pass


def download_file(downloaded_file_name, url, max_connections=2, max_concurrent=5):
    """Download file to specified location."""
    from commandlib import Command, CommandError, run
    if os.path.exists(downloaded_file_name):
        return

    log("Downloading: {}\n".format(url))
    aria2c = Command("aria2c")
    aria2c = aria2c("--max-connection-per-server={}".format(max_connections))
    aria2c = aria2c("--max-concurrent-downloads={}".format(max_concurrent))

    try:
        if os.path.isabs(downloaded_file_name):
            run(aria2c("--dir=/", "--out={}.part".format(downloaded_file_name), url))
        else:
            run(aria2c("--dir=.", "--out={}.part".format(downloaded_file_name), url))
    except CommandError:
        raise DownloadError("Failed to download {}. Re-running may fix the problem.".format(url))

    shutil.move(downloaded_file_name + ".part", downloaded_file_name)


def get_hitch_directory():
    """Get the hitch directory by working backwards from the virtualenv python."""
    import sys
    return os.path.abspath(
        os.path.join(os.path.dirname(sys.executable), "..", "..")
    )



from subprocess import check_call
from os import path, makedirs, listdir, sep


def snapshot(source, destination, name=None):
    """Snapshot one directory to another. Specify names to snapshot small, named differences."""
    source = source + sep
    destination = destination + sep

    if not path.isdir(source):
        raise RuntimeError("source is not a directory")

    if path.exists(destination):
        if not path.isdir(destination):
            raise RuntimeError("destination is not a directory")

        if name is None:
            raise RuntimeError("can't snapshot base snapshot if destination exists")

    snapdir = path.join(destination, ".snapdir")
    
    if path.exists(path.join(source, ".snapdir")):
        raise RuntimeError("snapdir exists in source directory")

    if name is None:
        check_call(["rsync", "--del", "-av", source, destination])
        makedirs(snapdir)
    else:
        if not path.exists(snapdir):
            raise RuntimeError("No snapdir in destination directory")

        check_call(["rsync", "--del", "-av", "--only-write-batch={}".format(path.join(snapdir, name)), source, destination])


def restore(destination, source, name=None):
    """Restore snapshots from one directory to another."""
    source = source + sep
    destination = destination + sep

    snapdir = path.join(destination, ".snapdir")

    check_call(["rsync", "--progress", "--del", "--exclude=.snapdir", "-av", destination, source])

    if name is not None:
        check_call(["rsync", "--del", "--exclude=.snapdir", "-av", "--read-batch={}".format(path.join(snapdir, name)), destination, source])



from jinja2.environment import Environment
from jinja2 import FileSystemLoader
from hitchtest.hitch_stacktrace import TestPosition
from os import path
import colorama
import json
import sys

TEMPLATE_DIR = path.join(path.dirname(path.realpath(__file__)), "templates")

class Results(object):
    def __init__(self, result_list, failedfast, colorless):
        self.result_list = result_list
        self.failedfast = failedfast
        self.colorless = colorless

    def failures(self):
        return [result for result in self.result_list if result.failure]

    def to_dict(self):
        return {
            'one_test': len(self.result_list) == 1,
            'at_least_one_failure': len(self.failures()) > 0,
            'total': len(self.result_list),
            'total_passes': len(self.result_list) - len(self.failures()),
            'total_failures': len(self.failures()),
            'duration': sum([result.duration for result in self.result_list]),
            'failures': [result.to_dict() for result in self.failures()],
            'failedfast': self.failedfast,
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_template(self, template):
        if template is None:
            template = "default.jinja2"
        env = Environment()
        env.loader = FileSystemLoader(TEMPLATE_DIR)
        tmpl = env.get_template(path.basename(template))
        return tmpl.render(
            results=self.to_dict(),
            json=self.to_json(),
            Fore=colorama.Fore if not self.colorless else '',
            Back=colorama.Back if not self.colorless else '',
            Style=colorama.Style if not self.colorless else '',
            TestPosition=TestPosition,
        )



from hitchtest.environment.utils import version_comparison, return_code_zero, check_output_lines
from collections import Counter
import unixpackage
import struct
import psutil
import glob
import sys
import re
import os


class HitchEnvironmentException(Exception):
    pass


def freeports(required_ports):
    """Verifies that none of the required_ports are not in use. Raise exception if they are."""
    unusable_ports = []
    import socket
    for port in required_ports:
        socket_to_try = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_to_try.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            socket_to_try.bind(('', port))
            socket_to_try.listen(1)
        except socket.error as e:
            unusable_ports.append(str(port))
        finally:
            socket_to_try.close()

    if len(unusable_ports) > 0:
        raise HitchEnvironmentException(
            "Required network port(s): {} not usable.".format(', '.join(unusable_ports))
        )


def packages(package_list):
    """Verify that a list of unixpackage packages are installed."""
    if not unixpackage.packages_installed(package_list):
        raise HitchEnvironmentException((
            "The following packages are required: {}.\n"
            "The command to install them on this environment is: {}.\n"
        ).format(', '.join(package_list), unixpackage.install_command(package_list)))


def debs(packages):
    """Verify that a list of debs are installed. Ignore if not a deb based system."""
    if len(packages) > 10:
        debs(packages[10:])
        packages = packages[:10]
    if return_code_zero(["which", "dpkg"]):
        if not return_code_zero(["dpkg", "--list",] + packages):
            raise HitchEnvironmentException(
                "sudo apt-get install {} : required for test to run".format(' '.join(packages))
            )


def brew(packages):
    """Verify that a list of brew packages are installed. Ignore if not a mac."""
    if sys.platform == "darwin":
        version_list = [
            x for x in check_output_lines(["brew", "list", "--versions"] + packages) if x != ""
        ]
        if len(version_list) < len(packages):
            raise HitchEnvironmentException(
                "brew install {} : required for test to run".format(' '.join(packages))
            )


def systembits(bits):
    """Verify that a system is 64 or 32 bit."""
    if not bits == struct.calcsize("P") * 8:
        raise HitchEnvironmentException(
            "{} bit system required to run test. This system is {} bit".format(str(bits))
        )


def internet_detected_after(timeout):
    """Verify that a system is connected to the internet."""
    if not return_code_zero(["ping", "-c", "1", "-W", str(timeout), "8.8.8.8"]):
        raise HitchEnvironmentException(
            "No internet detected after {} seconds. Ping failed.".format(timeout)
        )


def approved_platforms(platforms):
    """Verify that the test is running on an approved platform."""
    if not sys.platform in platforms:
        raise HitchEnvironmentException(
            "This platform is {}. This test will only run on {}.".format(' '.format(platforms))
        )


def x_available(true_or_false):
    """Checks if DISPLAY enviroment variable is set, hence that X is available."""
    if "DISPLAY" not in os.environ:
        raise HitchEnvironmentException(
            """The DISPLAY environment variable is not set. See """
            """https://stackoverflow.com/questions/784404/how-can-i-specify-a-display """
            """for more details about what to do to fix this issue."""
        )


def i_am_root(true_or_false):
    """Verify that the user running this test either is, or is not, root."""
    import getpass
    if getpass.getuser() == "root" and not true_or_false:
        raise HitchEnvironmentException(
            """The current user is root. You must run hitch clean and then """
            """try running the test(s) again as a non-root user."""
        )

    if getpass.getuser() != "root" and true_or_false:
        raise HitchEnvironmentException(
            """The current user is *not* root. You must run hitch clean and then """
            """try running the test(s) again as the root user."""
        )


def need_free_inotify_watches(number_of_watches_needed):
    """
    Verify that enough inotify watches are available. Raise exception if they are not.

    Only valid on linux, so skipped on non-linux platforms.
    """
    from path import Path
    if "linux" in sys.platform:
        max_user_watches = int(
            Path("/proc/sys/fs/inotify/max_user_watches").bytes().decode('utf8').replace('\n', '')
        )

        existing_user_watches = len([
            fulllink for fulllink in (os.readlink(x) for x in glob.glob("/proc/*/fd/*")
                if os.path.exists(x)) if "inotify" in fulllink
        ])

        if max_user_watches - existing_user_watches < number_of_watches_needed:
            raise HitchEnvironmentException((
                "You must increase max_user_watches in order to run your tests.\n\n"
                "Number of inotify watches required: {}\n"
                "Number of inotify watches left: {}\n"
                "Max user watches: {}\n"
                "How to temporarily raise number of watches:\n\n"
                "  sudo echo 16384 > /proc/sys/fs/inotify/max_user_watches\n\n"
                "How to permanently raise the number of watches:\n\n"
                "  Add the line 'fs.inotify.max_user_watches=16384' to the end of /etc/sysctl.conf"
            ).format(
                number_of_watches_needed,
                max_user_watches - existing_user_watches,
                max_user_watches,
            ))


# TODO : rpm package availability.
# TODO : arch package availability.
# TODO : all other package managers
# TODO : Kernel SHMMAX, glibc version, CPU usage,
# TODO : psutil.phymem_usage().available / 1024 / 1024
# TODO : psutil.cpu_percent()
# TODO : psutil.cpu_count()
# TODO : Linux Kernel version
# TODO : Mac OS X kernel version
# TODO : ulimit
# TODO : nslookup -timeout=1 microsoft.com


#def linux_distro(distros):
    #if path.exists("/etc/issue"):
        #with open("/etc/issue", "r") as issue_handle:
            #issue = issue_handle.read()

        #if "ubuntu" in distros and issue.startswith("Ubuntu"):
            #return True
        #if "redhat" in distros and issue.startswith("Red Hat"):
            #return True
        #return False
    #return True

#def ubuntu_version(version_check):
    #if path.exists("/etc/issue"):
        #with open("/etc/issue", "r") as issue_handle:
            #issue = issue_handle.read()
        #version = re.search("[0-9]+\.[0-9]+\.?[0-9]*", issue).group(0)
        #return version_comparison(version_check, version)
    #return True



from commandlib import Commands
from os import path, makedirs
import sys


class PackageNotBuilt(Exception):
    pass

class PackageVerificationFailure(Exception):
    pass

class InvalidPackageVersion(Exception):
    pass


class HitchPackage(object):
    name = None

    def __init__(self):

        # 3 directories above virtualenv python binary is the hitchdir
        self.hitch_package_directory = path.abspath(
            path.join(path.expanduser("~"), ".hitchpkg")
        )

        if self.name is None:
            raise PackageVerificationFailure("This package does not have self.name set!")

    def get_build_directory(self):
        if not path.exists(self.hitch_package_directory):
            makedirs(self.hitch_package_directory)
        return self.hitch_package_directory

    def get_downloads_directory(self):
        if not path.exists(self.hitch_package_directory):
            makedirs(self.hitch_package_directory)
        return self.hitch_package_directory

    def check_version(self, version, versions_list, issues_url, name=None):
        package_name = self.name if name is None else name
        if str(version) not in versions_list:
            raise InvalidPackageVersion(
                "{} version {} not in list of approved versions: \n{}\n"
                "Raise a ticket at {} "
                "if you think it should be.".format(
                    package_name, version, versions_list, issues_url
                )
            )
        return version

    def download_file(self, name, url):
        pass

    def extract_archive(self, filename, destination):
        pass

    def build(self):
        pass

    def verify(self):
        pass
    
    @property
    def cmd(self):
        return Commands(self.bin_directory)



"""Command line interface to hitchtest."""
from click import command, group, argument, option
from hitchtest.utils import log, warn, signals_trigger_exit
from sys import exit, executable
from os import path, walk, chdir
from hitchtest.settings import Settings
from hitchtest import module
from hitchtest import suite
import yaml as pyyaml
import fnmatch
import signal
import json


@command()
@argument('filenames', nargs=-1)
@option('-y', '--yaml', is_flag=True, help='Output the YAML tests (for debugging).')
@option('-q', '--quiet', is_flag=True, help='Quiet mode. Do not print test output to screen. (DEPRECATED)')
@option('-t', '--tags', default=None, help='Only run tests with comma-separated specified tag(s). (e.g. --tags tag-1,tag-2,tag.3)')
@option('-s', '--settings', default=None, help="Load settings from specified file.")
@option('-e', '--extra', default=None, help="""Load extra settings on command line as JSON (e.g. --extra '{"postgres_version": "3.5.5"}'""")
def cli(filenames, yaml, quiet, tags, settings, extra):
    """Run test files or entire directories containing .test files."""
    # .hitch/virtualenv/bin/python <- 4 directories up from where the python exec resides
    engine_folder = path.abspath(path.join(executable, "..", "..", "..", ".."))
    filenames = [path.abspath(path.relpath(filename, engine_folder)) for filename in filenames]
    chdir(engine_folder)

    if quiet:
        warn((
            "The --quiet switch has been deprecated. You can make your tests run quietly "
            "by setting the property quiet to True via --extra or in a settings file.\n\n"
            "See here for more details on the change : \n"
            "https://hitchtest.readthedocs.org/en/latest/faq/why_was_hitch_behavior_changed.html\n"
        ))
        exit(1)

    #new_default_settings_filename = path.join(engine_folder, "all.settings")
    #settings_filename = None if settings is None else path.join(engine_folder, settings)

    #if path.exists(new_default_settings_filename):
        #_overwrite_settings_with_yaml(new_default_settings_filename, settings_dict)

    #if settings_filename is not None:
        #if not path.exists(settings_filename):
            #warn("Settings file '{}' could not be found!\n".format(settings_filename))
            #exit(1)
        #else:
            ## Load settings from specified file, if it exists
            #_overwrite_settings_with_yaml(settings_filename, settings_dict)

    ## Load extra settings from command line JSON and overwrite what's already set
    #if extra is not None:
        #try:
            #settings_dict.update(json.loads(extra).items())
        #except ValueError as error:
            #warn("""{} in:\n ==> --extra '{}' (must be valid JSON)\n""".format(str(error), extra))
            #exit(1)
        #except AttributeError:
            #warn("""Error in:\n ==> --extra '{}' (must be valid JSON and not a list)\n""".format(extra))
            #exit(1)

    settings_dict = Settings(engine_folder, settings, extra)
    settings_dict['engine_folder'] = engine_folder
    if 'quiet' not in settings_dict:
        settings_dict['quiet'] = False

    if len(filenames) == 0:
        warn("No tests specified.\n")
        exit(1)

    # Get list of files from specified files/directories
    matches = []
    test_not_found = False
    for filename in filenames:
        if not path.exists(filename):
            warn("Test '{}' not found.\n".format(filename))
            test_not_found = True
        if path.isdir(filename):
            for root, dirnames, filenames_in_dir in walk(filename):
                for filename_in_dir in fnmatch.filter(filenames_in_dir, '*.test'):
                    if ".hitch" not in root: # Ignore everything in .hitch
                        matches.append(path.join(root, filename_in_dir))
        else:
            matches.append(filename)

    if test_not_found:
        exit(1)

    # Get list of modules from matching directly specified files from command line
    # and indirectly (in the directories of) directories specified from cmd line
    test_modules = []
    for filename in matches:
        if filename.endswith(".test"):
            test_modules.append(module.Module(filename, settings_dict))
        else:
            warn(
                "Tests must have the extension .test"
                "- '{}' doesn't have that extension\n".format(filename)
            )
            exit(1)

    test_suite = suite.Suite(test_modules, settings_dict, tags)

    if yaml:
        test_suite.printyaml()
    else:
        returned_results = test_suite.run(quiet=quiet)

        # Lines must be split to prevent stdout blocking
        result_lines = returned_results.to_template(
            template=settings_dict.get('results_template', None)
        ).split('\n')

        for line in result_lines:
            log("{}\n".format(line))

        exit(1 if len(returned_results.failures()) > 0 else 0)

def run():
    """Run hitch tests"""
    signals_trigger_exit()
    cli()

if __name__ == '__main__':
    run()



from os import path, makedirs
from hitchtest.utils import get_hitch_directory
from path import Path
import json


HOW_TO_USE_MODIFICATION_CHECKER = """\
You cannot treat monitor as a boolean.

It must be used as a context manager like so:

from hitchtest import monitor

with monitor([filename1, filename2...]) as changed:
    if changed:
        # run command
    else:
        # don't run command
"""


class Changed(object):
    """
    Object representation of what the monitor has changed.
    
    len(changes) == number of files changed
    if changes: -> True if at least one file has changed
    
    for change in changes:
        print(change)
    """
    def __init__(self, filenames):
        self._filenames = filenames
    
    def __len__(self):
        return len(self._filenames)
    
    def __getitem__(self, index):
        return Path(self._filenames[index])
    
    def __nonzero__(self):
        return len(self._filenames) > 0
    
    def __bool__(self):
        return len(self._filenames) > 0
    
    


class monitor(object):
    def __init__(self, includes, excludes=None):
        self.includes = [includes, ] if type(includes) is str else includes
        self.excludes = excludes if excludes is not None else []
        self.modififications_filename = path.join(get_hitch_directory(), "modified.json")
    
    def __enter__(self):
        # Get all existing modification timestamps
        if path.exists(self.modififications_filename):
            with open(self.modififications_filename, 'r') as handle:
                self.modifications = json.loads(handle.read())
        else:
            self.modifications = {}

        list_of_modified_filenames = []

        # Total list of filenames
        filenames = list(set(self.includes) - set(self.excludes))

        # Check if any of the files have been updated since
        for relfilename in filenames:
            absfilename = path.abspath(relfilename)
            modified_time = path.getmtime(absfilename)
            
            if absfilename in self.modifications:
                if modified_time > self.modifications[absfilename]:
                    self.modifications[absfilename] = modified_time
                    list_of_modified_filenames.append(absfilename)
            else:
                self.modifications[absfilename] = modified_time
                list_of_modified_filenames.append(absfilename)

        return Changed(list_of_modified_filenames)

    def __exit__(self, type, value, traceback):
        # If an exception is thrown, don't update modifications
        if traceback is None:
            # Check for modifications by nested modification checkers
            if path.exists(self.modififications_filename):
                with open(self.modififications_filename, 'r') as handle:
                    updated_modifications = json.loads(handle.read())
            else:
                updated_modifications = {}

            # If a modification has been updated since __enter__, don't overwrite it
            for filename in updated_modifications:
                if updated_modifications[filename] > self.modifications[filename]:
                    self.modifications[filename] = updated_modifications[filename]
            
            with open(self.modififications_filename, 'w') as handle:
                handle.write(json.dumps(self.modifications))

    def __bool__(self):
        raise NotImplementedError(HOW_TO_USE_MODIFICATION_CHECKER)

    def __nonzero__(self):
        raise NotImplementedError(HOW_TO_USE_MODIFICATION_CHECKER)
    
is_modified = monitor


import inspect
import signal
from hitchtest import utils
from hitchtest.signal_manager import SignalManager
from path import Path


class HitchAbortedException(Exception):
    pass


class HitchPathError(Exception):
    pass


#class HitchStatePath(Path):
    #def snapshot(path, name=None):
        ##self.abspath()
    
    #def restore(path, name=None):
        ##self.abspath()


class Paths(object):
    def __init__(self):
        self.hitch = Path(utils.get_hitch_directory())
        self.engine = self.hitch.parent
        self.state = self.hitch.joinpath("s")
        
        if not self.state.exists():
            self.state.mkdir()
        
    def __setattr__(self, name, value):
        if not isinstance(value, Path):
            raise HitchPathError("Path '{}' must be of type pathlib.Path".format(value))
        self.__dict__[name] = value


class ExecutionEngine(object):
    def __init__(self, settings, preconditions):
        self.settings = settings
        self.preconditions = preconditions
        self.signal_manager = SignalManager(self.abort)
        self.path = Paths()

    def set_up(self):
        pass

    def on_success(self):
        pass

    def on_failure(self):
        pass

    def tear_down(self):
        pass

    def ipython(self, message=None):
        self.signal_manager.attach_handler(signal.SIGINT, signal.default_int_handler)
        utils.ipython_embed(message)
        self.signal_manager.attach_handler(signal.SIGINT, self.abort)

    def abort(self, signal, frame):
        self.aborted = True
        utils.stop_ipython()
        raise HitchAbortedException



from hitchtest.utils import warn
from os.path import join, exists
from sys import exit
import yaml as pyyaml
import json


class Settings(object):
    def __init__(self, engine_directory, override_settings_filename, extra_string):
        
        if exists(join(engine_directory, "settings.yml")):
            warn((
                "settings.yml as a 'default settings' file has been deprecated. "
                "Rename to all.settings instead:\n\n"
                "See here for more details on the change : \n"
                "https://hitchtest.readthedocs.org/en/latest/faq/why_was_hitch_behavior_changed.html\n"
            ))
            exit(1)

        self.settings_dict = {}
        
        full_override_settings_filename = None if override_settings_filename is None else join(engine_directory, override_settings_filename)

        if exists(join(engine_directory, "all.settings")):
            self._overwrite_settings_with_yaml(join(engine_directory, "all.settings"))

        if override_settings_filename is not None:
            if not exists(override_settings_filename):
                warn("Settings file '{}' could not be found!\n".format(override_settings_filename))
                exit(1)
            else:
                # Load settings from specified file, if it exists
                self._overwrite_settings_with_yaml(override_settings_filename)

        # Load extra settings from command line JSON and overwrite what's already set
        if extra_string is not None:
            try:
                self.settings_dict.update(json.loads(extra_string).items())
            except ValueError as error:
                warn("""{} in:\n ==> --extra '{}' (must be valid JSON)\n""".format(str(error), extra_string))
                exit(1)
            except AttributeError:
                warn("""Error in:\n ==> --extra '{}' (must be valid JSON and not a list)\n""".format(extra_string))
                exit(1)

    def __getitem__(self, key):
        if key not in self.settings_dict:
            raise IndexError
        return self.settings_dict[key]

    def get(self, key, default_value=None):
        if key not in self.settings_dict:
            return default_value
        else:
            return self.settings_dict[key]

    def __setitem__(self, key, value):
        self.settings_dict[key] = value

    def as_dict(self):
        return self.settings_dict

    def _overwrite_settings_with_yaml(self, settings_filename):
        """Load YAML and overwrite """
        if exists(settings_filename):
            with open(settings_filename) as settingsfile_handle:
                settingsfile_contents = settingsfile_handle.read()

            try:
                self.settings_dict.update(pyyaml.load(settingsfile_contents))
            except pyyaml.parser.MarkedYAMLError as error:
                warn("YAML parser error in {}:\n{}\nError:{}\n".format(
                    settings_filename, settingsfile_contents, str(error),
                ))
                exit(1)
            except ValueError:
                warn("YAML parser error in {}. Should be associative array not list:\n\n{}\n".format(
                    settings_filename, settingsfile_contents,
                ))
                exit(1)



from hitchtest.arguments import Arguments
from hitchtest import utils
import yaml


class StepNotFound(Exception):
    pass

class Step(object):
    def __init__(self, yaml_step, index, underscore_case_steps=True):
        self.index = int(index)
        self.underscore_case_steps = underscore_case_steps
        if type(yaml_step) is str:
            self.name = str(yaml_step)
            self.arguments = Arguments(None)
        elif type(yaml_step) is dict and len(yaml_step.keys()) == 1:
            self.name = list(yaml_step.keys())[0]
            self.arguments = Arguments(list(yaml_step.values())[0])
        else:
            raise RuntimeError("Invalid YAML in step '{}'".format(yaml_step))

    def __str__(self):
        return self.name

    def underscore_case_name(self):
        return utils.to_underscore_style(self.name)

    def as_yaml(self):
        if self.arguments.is_none:
            return yaml.dump([self.name], default_flow_style=False).rstrip()
        else:
            return yaml.dump([{self.name: self.arguments.to_dict(), }], default_flow_style=False).rstrip()

    def to_dict(self):
        return {
            'index': self.index,
            'name': self.name,
            'arguments': self.arguments.to_dict(),
            'yaml': self.as_yaml(),
        }

    def run(self, engine):
        if hasattr(engine, self.underscore_case_name()):
            if self.arguments.is_none:
                getattr(engine, self.underscore_case_name())()
            elif self.arguments.single_argument:
                getattr(engine, self.underscore_case_name())(self.arguments.argument)
            else:
                if self.underscore_case_steps:
                    getattr(engine, self.underscore_case_name())(**self.arguments.pythonized_kwargs())
                else:
                    getattr(engine, self.underscore_case_name())(**self.arguments.kwargs)
        else:
            raise StepNotFound("Step with name '{}' not found in execution engine.".format(self.underscore_case_name()))



from hitchtest import utils


class Arguments(object):
    """A null-argument, single argument or group of arguments of a hitch test step."""

    def __init__(self, yaml_args):
        """Create arguments from dict (from yaml)."""
        if yaml_args is None:
            self.is_none = True
            self.single_argument = False
        elif type(yaml_args) is str or type(yaml_args) is int or type(yaml_args) is float or type(yaml_args) is bool:
            self.is_none = False
            self.single_argument = True
            self.argument = yaml_args
        else:
            self.is_none = False
            self.single_argument = False
            self.kwargs = yaml_args

    def pythonized_kwargs(self):
        pythonized_dict = {}
        for key, value in self.kwargs.items():
            pythonized_dict[utils.to_underscore_style(key)] = value
        return pythonized_dict

    def to_dict(self):
        if self.is_none:
            return None
        elif self.single_argument:
            return self.argument
        else:
            return self.kwargs



class Result(object):
    def __init__(self, test, failure, duration, stacktrace=None, aborted=False, test_out=None, test_err=None):
        self.test = test.to_dict()
        self.failure = failure
        self.duration = duration
        self.stacktrace = stacktrace
        self.aborted = aborted
        self.test_out = test_out
        self.test_err = test_err

    def to_dict(self):
        return {
            'test': self.test,
            'failure': self.failure,
            'duration': self.duration,
            'stacktrace': self.stacktrace,
        }
